from typing import Dict, List, Set, Union
from collections import defaultdict, Counter
import re

from fuzzy_search.fuzzy_string import SkipGram, text2skipgrams


def is_valid_label(label: Union[str, List[str]]) -> bool:
    """Test whether label has a valid value.

    :param label: a phrase label (either a string or a list of strings)
    :type label: Union[str, List[str]]
    :return: whether the label is valid
    :rtype: bool
    """
    if isinstance(label, list):
        for item in label:
            if not isinstance(item, str):
                return False
        return True
    return isinstance(label, str)


class Phrase(object):

    def __init__(self, phrase: Union[str, Dict[str, str]], ngram_size: int = 2, skip_size: int = 2,
                 early_threshold: int = 3, late_threshold: int = 3, within_range_threshold: int = 3,
                 ignore_case: bool = False):
        if isinstance(phrase, str):
            phrase = {"phrase": phrase}
        self.name = phrase["phrase"]
        self.phrase_string = self.name if not ignore_case else self.name.lower()
        self.exact_string = re.escape(self.phrase_string)
        self.extact_word_boundary_string = re.compile(rf"\b{self.exact_string}\b")
        self.label = None
        self.max_offset: int = -1
        self.max_end: int = -1
        self.label_set: Set[str] = set()
        self.label_list: List[str] = []
        self.properties = phrase
        self.ngram_size = ngram_size
        self.skip_size = skip_size
        self.early_threshold = early_threshold
        self.late_threshold = len(self.name) - late_threshold - ngram_size
        self.within_range_threshold = within_range_threshold
        self.ignore_case = ignore_case
        self.skipgrams = [skipgram for skipgram in text2skipgrams(self.phrase_string,
                                                                  ngram_size=ngram_size, skip_size=skip_size)]
        self.skipgram_set = set([skipgram.string for skipgram in self. skipgrams])
        self.skipgram_index: Dict[str, List[SkipGram]] = defaultdict(list)
        self.skipgram_index_lower: Dict[str, List[SkipGram]] = defaultdict(list)
        self.skipgram_freq = Counter([skipgram.string for skipgram in self.skipgrams])
        self.early_skipgram_index = {skipgram.string: skipgram for skipgram in
                                     self.skipgrams if skipgram.offset < early_threshold}
        self.late_skipgram_index = {skipgram.string: skipgram for skipgram in
                                    self.skipgrams if skipgram.offset > self.late_threshold}
        # add lowercase version to allow both matching with and without ignore_case
        self.skipgrams_lower = [skipgram for skipgram in text2skipgrams(self.phrase_string.lower(),
                                                                        ngram_size=ngram_size, skip_size=skip_size)]
        self.early_skipgram_index_lower = {skipgram: skipgram for skipgram in self.skipgrams_lower
                                           if skipgram.offset < early_threshold}
        self.late_skipgram_index_lower = {skipgram.string: skipgram for skipgram in self.skipgrams_lower
                                          if skipgram.offset > self.late_threshold}
        self.skipgram_freq_lower = Counter([skipgram.string for skipgram in self.skipgrams_lower])
        self.num_skipgrams = len(self.skipgrams)
        self.skipgram_distance = {}
        self.metadata: dict = phrase
        self.words: List[str] = [word for word in re.split(r"\W+", self.phrase_string) if word != ""]
        self.word_set: Set[str] = set(self.words)
        self.first_word = None if len(self.words) == 0 else self.words[0]
        self.last_word = None if len(self.words) == 0 else self.words[-1]
        self.num_words = len(self.words)
        if "label" in phrase:
            self.set_label(phrase["label"])
        if len(phrase.keys()) > 1:
            self.add_metadata(phrase)
        self._index_skipgrams()
        self._set_within_range()

    def __repr__(self):
        return f"Phrase({self.phrase_string}, {self.label})"

    # internal methods

    def _index_skipgrams(self) -> None:
        """Turn the phrase into a list of skipgrams and index them with their offset(s) as values."""
        for skipgram in self.skipgrams:
            self.skipgram_index[skipgram.string] += [skipgram]
        for skipgram in self.skipgrams_lower:
            self.skipgram_index_lower[skipgram.string] += [skipgram]

    def _set_within_range(self):
        self.skipgram_distance = {}
        for index1 in range(0, len(self.skipgrams)-1):
            skipgram1 = self.skipgrams[index1]
            for index2 in range(index1+1, len(self.skipgrams)):
                skipgram2 = self.skipgrams[index2]
                if skipgram2.offset - skipgram1.offset > self.within_range_threshold:
                    continue
                if (skipgram1, skipgram2) not in self.skipgram_distance:
                    self.skipgram_distance[(skipgram1, skipgram2)] = skipgram2.offset - skipgram1.offset
                elif self.skipgram_distance[(skipgram1, skipgram2)] > skipgram2.offset - skipgram1.offset:
                    self.skipgram_distance[(skipgram1, skipgram2)] = skipgram2.offset - skipgram1.offset

    # external methods

    def set_label(self, label: Union[str, List[str]]) -> None:
        """Set the label(s) of a phrase. Labels must be string and can be a single string or a list.

        :param label: the label(s) of a phrase
        :type label: Union[str, List[str]]
        """
        if not is_valid_label(label):
            raise ValueError("phrase label must be a single string or a list of strings:", label)
        self.label = label
        if isinstance(label, str):
            self.label_set = {label}
            self.label_list = [label]
        else:
            self.label_set = set(label)
            self.label_list = label

    def has_label(self, label_string: str) -> bool:
        """Check if a given label belongs to at least one phrase in the phrase model.

        :param label_string: a label string
        :type label_string: str
        :return: a boolean whether the label is part of the phrase model
        :rtype: bool
        """
        if isinstance(self.label, list):
            return label_string in self.label
        else:
            return label_string == self.label

    def add_metadata(self, metadata_dict: Dict[str, any]) -> None:
        """Add key/value pairs as metadata for this phrase.

        :param metadata_dict: a dictionary of key/value pairs as metadata
        :type metadata_dict: Dict[str, any]
        :return: None
        :rtype: None
        """
        for key in metadata_dict:
            self.metadata[key] = metadata_dict[key]
            if key == "label":
                self.set_label(metadata_dict[key])
            elif key == "max_offset":
                self.add_max_offset(metadata_dict["max_offset"])

    def add_max_offset(self, max_offset: int) -> None:
        """Add a maximum offset for matching a phrase in a text.

        :param max_offset: the maximum offset to allow a phrase to match
        :type max_offset: int
        """
        if not isinstance(max_offset, int):
            raise TypeError("max_offset must be a positive integer")
        if max_offset < 0:
            raise ValueError("max_offset must be positive")
        self.max_offset = max_offset
        self.max_end = self.max_offset + len(self.phrase_string)

    def has_skipgram(self, skipgram: str) -> bool:
        """For a given skipgram, return boolean whether it is in the index

        :param skipgram: an skipgram string
        :type skipgram: str
        :return: A boolean whether skipgram is in the index
        :rtype: bool"""
        return skipgram in self.skipgram_index.keys()

    def skipgram_offsets(self, skipgram_string: str) -> Union[None, List[int]]:
        """For a given skipgram return the list of offsets at which it appears.

        :param skipgram_string: an skipgram string
        :type skipgram_string: str
        :return: A list of string offsets at which the skipgram appears
        :rtype: Union[None, List[int]]"""
        if not self.has_skipgram(skipgram_string):
            return None
        return [skipgram.offset for skipgram in self.skipgram_index[skipgram_string]]

    def within_range(self, skipgram1, skipgram2):
        if not self.has_skipgram(skipgram1) or not self.has_skipgram(skipgram2):
            return False
        elif (skipgram1, skipgram2) not in self.skipgram_distance:
            return False
        elif self.skipgram_distance[(skipgram1, skipgram2)] > self.within_range_threshold:
            return False
        else:
            return True

    def is_early_skipgram(self, skipgram: str) -> bool:
        """For a given skipgram, return boolean whether it appears early in the phrase.

        :param skipgram: an skipgram string
        :type skipgram: str
        :return: A boolean whether skipgram appears early in the phrase
        :rtype: bool"""
        return skipgram in self.early_skipgram_index



from unittest import TestCase

from fuzzy_search.fuzzy_template import FuzzyTemplate, FuzzyTemplateGroupElement, FuzzyTemplateLabelElement
from fuzzy_search.fuzzy_phrase import Phrase
from fuzzy_search.fuzzy_phrase_model import PhraseModel
from data.demo_data import DemoData


class TestFuzzyTemplateElement(TestCase):

    def test_template_accepts_label_only(self):
        template = FuzzyTemplateLabelElement(label="some_label")
        self.assertEqual(template.label, "some_label")

    def test_template_accepts_label_and_cardinality(self):
        template = FuzzyTemplateLabelElement(label="some_label", cardinality="multi")
        self.assertEqual(template.cardinality, "multi")

    def test_template_rejects_invalid_cardinality_value(self):
        error = None
        try:
            FuzzyTemplateLabelElement(label="some_label", cardinality="more_than_one")
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)


class TestFuzzyTemplateGroup(TestCase):

    def setUp(self) -> None:
        self.element = FuzzyTemplateLabelElement(label="element_label")

    def test_template_group_accepts_label_only(self):
        template_group = FuzzyTemplateGroupElement([self.element], label="group_label")
        self.assertEqual(template_group.label, "group_label")

    def test_template_group_accepts_label_and_order(self):
        template_group = FuzzyTemplateGroupElement([self.element], label="group_label")
        self.assertEqual(template_group.ordered, True)


class TestFuzzyTemplate(TestCase):

    def setUp(self) -> None:
        phrase_model = [{"phrase": "test phrase", "label": ["test_label"]}]
        self.phrase_model = PhraseModel(model=phrase_model)

    def test_template_generation(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model, template_json={"label": "test_label"})
        self.assertEqual(isinstance(template, FuzzyTemplate), True)

    def test_template_can_get_phrase_by_label(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model, template_json={"label": "test_label"})
        phrases = template.get_label_phrases("test_label")
        self.assertEqual(len(phrases), 1)

    def test_template_get_phrase_by_label_returns_phrase_object(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model, template_json={"label": "test_label"})
        phrases = template.get_label_phrases("test_label")
        self.assertEqual(isinstance(phrases[0], Phrase), True)

    def test_template_get_phrase_by_label_returns_correct_phrase(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model, template_json={"label": "test_label"})
        phrases = template.get_label_phrases("test_label")
        self.assertEqual(phrases[0].phrase_string, "test phrase")

    def test_template_register_simple_element(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model,
                                 template_json={"label": "test_label", "extra": "extra"})
        self.assertEqual(template.has_label("test_label"), True)

    def test_template_register_simple_element_as_multi_if_no_cardinality(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model, template_json=[{"label": "test_label"}])
        self.assertEqual(template.label_element_index["test_label"].cardinality, "single")

    def test_template_register_simple_element_with_list_labels(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model,
                                 template_json=[{"label": "test_label", "required": True}])
        element = template.get_element("test_label")
        self.assertEqual(element.required, True)

    def test_template_cannot_register_element_with_unknown_label(self):
        error = None
        try:
            FuzzyTemplate(phrase_model=self.phrase_model, template_json=[{"label": "other_label", "required": True}])
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_template_can_ignore_element_with_unknown_label(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model,
                                 template_json=[{"label": "other_label", "required": True}],
                                 ignore_unknown=True)
        self.assertEqual(template.has_label("other_label"), False)

    def test_template_can_return_required_elements(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model, template_json=[{"label": "test_label"}])
        self.assertEqual(len(template.get_required_elements()), 0)

    def test_template_returns_all_required_elements(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model,
                                 template_json=[{"label": "test_label", "required": True}])
        self.assertEqual(len(template.get_required_elements()), 1)

    def test_template_returns_all_required_element_labels(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model,
                                 template_json=[{"label": "test_label", "required": True}])
        self.assertEqual(len(template.get_required_labels()), 1)
        self.assertEqual("test_label" in template.get_required_labels(), True)

    def test_template_can_register_group_elements(self):
        template = FuzzyTemplate(phrase_model=self.phrase_model,
                                 template_json=[{"type": "group", "label": "group_label", "elements": ["test_label"]}])
        self.assertEqual(template.has_group("group_label"), True)
        self.assertEqual("group_label" in template.group_element_index, True)


class TestFuzzyTemplateWithRealData(TestCase):

    def setUp(self) -> None:
        demo_data = DemoData()
        self.auction_data = demo_data.get_dataset("auction_advertisements")

    def test_template_can_read_real_data(self):
        phrase_model = PhraseModel(model=self.auction_data["phrases"])
        template = FuzzyTemplate(phrase_model=phrase_model,
                                 template_json=self.auction_data["template"],
                                 ignore_unknown=True)
        self.assertEqual(template.has_label("product_unit"), True)



from unittest import TestCase
from fuzzy_search.fuzzy_phrase import Phrase
from fuzzy_search.fuzzy_phrase_model import PhraseModel
from fuzzy_search.fuzzy_string import SkipGram
from fuzzy_search.fuzzy_phrase_searcher import FuzzyPhraseSearcher, SkipMatches, Candidate
from fuzzy_search.fuzzy_phrase_searcher import filter_skipgram_threshold, get_skipmatch_candidates

class TestSkipMatches(TestCase):

    def test_skip_matches_registers_match(self):
        skip_matches = SkipMatches(2, 2)
        phrase = Phrase('test')
        skipgram = SkipGram('ts', 0, 3)
        skip_matches.add_skip_match(skipgram, phrase)
        self.assertTrue(phrase in skip_matches.match_set)


class TestCandidate(TestCase):

    def test_candidate_detects_no_match_with_no_skip_match(self):
        phrase = Phrase('test')
        candidate = Candidate(phrase)
        self.assertEqual(candidate.is_match(0.5), False)

    def test_candidate_detects_no_match(self):
        phrase = Phrase('test')
        candidate = Candidate(phrase)
        skipgram = SkipGram('ts', 0, 3)
        candidate.add_skip_match(skipgram)
        self.assertEqual(candidate.is_match(0.5), False)

    def test_candidate_has_skipgram_overlap(self):
        phrase = Phrase('test')
        candidate = Candidate(phrase)
        skipgram = SkipGram('ts', 0, 3)
        candidate.add_skip_match(skipgram)
        self.assertTrue(candidate.get_skip_set_overlap() > 0.0)


class TestFuzzyPhraseSearcher(TestCase):

    def test_can_make_default_phrase_searcher(self):
        searcher = FuzzyPhraseSearcher()
        self.assertNotEqual(searcher, None)

    def test_can_add_phrases_as_strings(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "test"
        searcher.index_phrases(phrases=[phrase])
        phrase_object = searcher.phrases.pop()
        self.assertEqual(phrase_object.phrase_string, phrase)

    def test_can_add_phrases_as_phrase_objects(self):
        searcher = FuzzyPhraseSearcher()
        phrase = Phrase("test")
        searcher.index_phrases(phrases=[phrase])
        self.assertTrue(phrase in searcher.phrases)

    def test_can_generate_skip_matches(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "test"
        searcher.index_phrases(phrases=[phrase])
        text = "this is a test"
        skip_matches = searcher.find_skipgram_matches({"text": text})
        phrase_object = searcher.phrases.pop()
        self.assertTrue(phrase_object in skip_matches.match_set)

    def test_can_filter_skipgram_threshold(self):
        searcher = FuzzyPhraseSearcher()
        phrase = Phrase("test")
        searcher.index_phrases(phrases=[phrase])
        text = "this is a test"
        skip_matches = searcher.find_skipgram_matches({"text": text})
        phrases = filter_skipgram_threshold(skip_matches, 0.5)
        self.assertEqual(len(phrases), 1)

    def test_can_get_candidates(self):
        searcher = FuzzyPhraseSearcher()
        phrase_model = PhraseModel(phrases=["test"])
        searcher.index_phrase_model(phrase_model=phrase_model)
        text = {"text": "this is a test"}
        skip_matches = searcher.find_skipgram_matches(text)
        phrases = get_skipmatch_candidates(text, skip_matches, 0.5, phrase_model=phrase_model)
        self.assertEqual(len(phrases), 1)

    def test_finds_multiple_candidates(self):
        searcher = FuzzyPhraseSearcher()
        phrase_model = PhraseModel(phrases=["test"])
        searcher.index_phrase_model(phrase_model=phrase_model)
        text = {"text": "a test is a test is a test"}
        skip_matches = searcher.find_skipgram_matches(text)
        phrases = get_skipmatch_candidates(text, skip_matches, 0.5, phrase_model=phrase_model)
        self.assertEqual(len(phrases), 3)

    def test_searcher_finds_near_match(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text consaint some typos."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_is_case_sensitive(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text CONSAINT some typos."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 0)

    def test_searcher_handles_ignorecase(self):
        searcher = FuzzyPhraseSearcher({"ignorecase": True})
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text CONSAINT some typos."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_uses_word_boundaries(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text containsi some typos."
        matches = searcher.find_matches(text)
        self.assertEqual(isinstance(matches, list), True)
        self.assertEqual("containsi", matches[0].string)

    def test_searcher_finds_repeat_phrases_as_multiple_matches(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text contains contains some repetition."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 2)
        self.assertEqual("contains", matches[0].string)
        self.assertEqual("contains", matches[1].string)

    def test_searcher_finds_correct_start(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text con contains some weirdness."
        matches = searcher.find_matches(text)
        self.assertEqual("contains", matches[0].string)

    def test_searcher_allows_length_variance(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "coffee"
        searcher.index_phrases(phrases=[phrase])
        text = "For sale two units of coffy."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_allows_length_variance_2(self):
        searcher = FuzzyPhraseSearcher()
        phrase = "Makelaars"
        searcher.index_phrases(phrases=[phrase])
        text = 'door de Alakei&ers by na gecompletecrt'
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_can_toggle_variants(self):
        searcher = FuzzyPhraseSearcher({"include_variants": True})
        self.assertEqual(searcher.include_variants, True)

    def test_searcher_can_register_variants(self):
        searcher = FuzzyPhraseSearcher({"include_variants": True})
        phrase = {"phrase": "okay", "variants": ["OK"]}
        searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        self.assertEqual(len(searcher.variants), 1)
        variant = searcher.variants.pop()
        self.assertEqual(variant.phrase_string, "OK")

    def test_searcher_can_match_variants(self):
        searcher = FuzzyPhraseSearcher({"include_variants": True})
        phrase = {"phrase": "okay", "variants": ["OK"]}
        searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        text = "This text is okay and this test is OK."
        matches = searcher.find_matches(text, include_variants=True)
        self.assertEqual(matches[1].phrase.phrase_string, phrase["phrase"])
        self.assertEqual(matches[1].variant.phrase_string, phrase["variants"][0])

    def test_searcher_can_toggle_distractors(self):
        searcher = FuzzyPhraseSearcher({"filter_distractors": True})
        self.assertEqual(searcher.filter_distractors, True)

    def test_searcher_can_register_distractors(self):
        searcher = FuzzyPhraseSearcher({"filter_distractors": True})
        phrase = {"phrase": "okay", "distractors": ["OK"]}
        searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        self.assertEqual(len(searcher.distractors), 1)
        distractor = searcher.distractors.pop()
        self.assertEqual(distractor.phrase_string, "OK")

    def test_searcher_can_match_distractors(self):
        searcher = FuzzyPhraseSearcher({"filter_distractors": True})
        phrase = {"phrase": "baking", "distractors": ["braking"]}
        searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        text = "This text is about baking and not about braking."
        matches = searcher.find_matches(text, filter_distractors=True)
        self.assertEqual(len(matches), 1)
    """


"""
class TestFuzzySearchExactMatch(TestCase):

    def setUp(self) -> None:
        self.searcher = FuzzyPhraseSearcher()
        self.phrase = {"phrase": "baking", "distractors": ["braking"]}
        self.searcher.index_phrase_model(phrase_model=PhraseModel([self.phrase]))

    def test_fuzzy_search_can_search_exact_match(self):
        text = "This text is about baking and not about braking."
        matches = self.searcher.find_exact_matches(text)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].string, "baking")

    def test_fuzzy_search_can_search_exact_match_with_word_boundaries(self):
        text = "This text is about baking and not about rebaking."
        matches = self.searcher.find_exact_matches(text, use_word_boundaries=True)
        self.assertEqual(len(matches), 1)

    def test_fuzzy_search_can_search_exact_match_without_word_boundaries(self):
        text = "This text is about baking and not about rebaking."
        matches = self.searcher.find_exact_matches(text, use_word_boundaries=False)
        self.assertEqual(len(matches), 2)
        self.assertEqual(matches[1].string, "baking")

    def test_fuzzy_search_can_search_exact_match_with_special_characters(self):
        searcher = FuzzyPhraseSearcher()
        phrase = {"phrase": "[baking]", "distractors": ["braking"]}
        searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        text = "This text is about [baking] and not about braking."
        matches = searcher.find_exact_matches(text, use_word_boundaries=False)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].string, "[baking]")

    def test_text_split(self):
        text = {
            "text": "Ntfangen een Missive van den Gouverneur Generaal van het eiland Amoras, "
                    + "verfoekende, dat dit beter getest moet worden.",
            "id": "urn:republic/inv=3825/meeting=1775-01-28/para=2"
        }
        phrases = [
            {"phrase": "ONtfangen een Missive van"},
            {"phrase": "Missive"},
            {"phrase": "Gouverneur Generaal"},
            {"phrase": "Gouverneur"},
            {"phrase": "Generaal van de"},
            {"phrase": "versoekende"}
        ]
        phrase_model = PhraseModel(model=phrases)
        searcher = FuzzyPhraseSearcher()
        searcher.index_phrase_model(phrase_model)
        exact_matches = searcher.find_exact_matches(text)
        self.assertEqual(len(exact_matches), 3)


class TestSearcherRealData(TestCase):

    def setUp(self) -> None:
        self.text1 = "ie Veucris den 5. Januaris 1725. PR&ASIDE, Den Heere Bentinck. PRASENTIEBUS, " + \
                     "De Heeren Jan Welderen , van Dam, Torck , met een extraordinaris Gedeputeerde" + \
                     " uyt de Provincie van Gelderlandt. Van Maasdam , vanden Boeizelaar , Raadtpen" + \
                     "fionaris van Hoornbeeck , met een extraordinaris Gedeputeerde uyt de Provinci" + \
                     "e van Hollandt ende Welt-Vrieslandt. Velters, Ockere , Noey; van Hoorn , met " + \
                     "een extraordinaris Gedeputeerde uyt de Provincie van Zeelandt. Van Renswoude " + \
                     ", van Voor{t. Van Schwartzenbergh, vander Waayen, Vegilin Van I{elmuden. Van " + \
                     "Iddekinge ‚ van Tamminga."
        self.text2 = "Mercuri: den 10. Jangarii, 1725. ia PRESIDE, Den Heere an Iddekinge. PRA&SENT" + \
                     "IBUS, De Heeren /an Welderen , van Dam, van Wynbergen, Torck, met een extraor" + \
                     "dinaris Gedeputeerde uyt de Provincie van Gelderland. Van Maasdam , Raadtpenf" + \
                     "ionaris van Hoorn=beeck. Velters, Ockerfe, Noey. Taats van Amerongen, van Ren" + \
                     "swoude. Vander Waasen , Vegilin, ’ Bentinck, van I(elmaden. Van Tamminga."
        self.config = {
            "char_match_threshold": 0.6,
            "ngram_threshold": 0.5,
            "levenshtein_threshold": 0.6,
            "ignorecase": False,
            "max_length_variance": 3,
            "ngram_size": 2,
            "skip_size": 2,
        }
        self.searcher = FuzzyPhraseSearcher(self.config)
        # create a list of domain phrases
        self.domain_phrases = [
            # terms for the chair and attendants of a meeting
            "PRAESIDE",
            "PRAESENTIBUS",
            # some weekdays in Latin
            "Veneris",
            "Mercurii",
            # some date phrase where any date in January 1725 should match
            "den .. Januarii 1725"
        ]
        self.phrase_model = PhraseModel(phrases=self.domain_phrases)
        # register the keywords with the searcher
        self.searcher.index_phrase_model(self.phrase_model)

    def test_fuzzy_search_text1_finds_four_matches(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(len(matches), 4)

    def test_fuzzy_search_text1_finds_friday(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[0].string, "Veucris")

    def test_fuzzy_search_text1_finds_date(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[1].string, "den 5. Januaris 1725")

    def test_fuzzy_search_text1_finds_president(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[2].string, "PR&ASIDE")

    def test_fuzzy_search_text1_finds_attendants(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[3].string, "PRASENTIEBUS")

    def test_fuzzy_search_text2_finds_four_matches(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(len(matches), 4)

    def test_fuzzy_search_text2_finds_friday(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[0].string, "Mercuri")

    def test_fuzzy_search_text2_finds_date(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[1].string, "den 10. Jangarii, 1725")

    def test_fuzzy_search_text2_finds_president(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[2].string, "PRESIDE")

    def test_fuzzy_search_text2_finds_attendants(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[3].string, "PRA&SENTIBUS")


class TestSearcherRealData2(TestCase):

    def setUp(self) -> None:
        self.text1 = 'TS gehoort het rapport van de Heeren I van Lynden'
        self.config = {
            "char_match_threshold": 0.7,
            "ngram_threshold": 0.7,
            "levenshtein_threshold": 0.7,
            "ignorecase": False,
            "include_variants": True,
            "max_length_variance": 3,
            "ngram_size": 2,
            "skip_size": 2,
        }
        self.searcher = FuzzyPhraseSearcher(self.config)
        # create a list of domain phrases
        self.domain_phrases = [
            {
                'phrase': 'den Heere',
                'variants': [
                    'de Heer',
                    'de Heeren',
                ]
            }
        ]
        self.phrase_model = PhraseModel(phrases=self.domain_phrases)
        # register the keywords with the searcher
        self.searcher.index_phrase_model(self.phrase_model)

    def test_searcher_find_no_overlapping_variants(self):
        phrase_matches = self.searcher.find_matches(self.text1)
        self.assertEqual(len(phrase_matches), 1)



from typing import Union

dutch_person_name_patterns = {
    "name_suffixes": r"( de jonge| de oude| junior| senior)",
    "name_pattern": r"(([A-Z](\w|-)+)( [A-Z](\w|-)+)*( van| de| der| den)*( [A-Z](\w|-)+)+( de jonge| de oude| junior| senior)?)",
    "name_comma_name_pattern": r"((([A-Z](\w|-)+)( [A-Z](\w|-)+)*( van| de| der| den)*( [A-Z](\w|-)+)+( de jonge| de oude| junior| senior)?), ?)+",
    "name_comma_pattern": r"(([A-Z](\w|-)+)( [A-Z](\w|-)+)*( van| de| der| den)*( [A-Z](\w|-)+)+( de jonge| de oude| junior| senior)?), ?",
}

dutch_date_patterns = {
    "week_day": r"(maandag|dinsdag|woensdag|donderdag|vrijdag|zaterdag|zondag)",
    "month": r"(jan(uari|.)?|feb(ruari|\.)?|maart|mrt|apr(il|\.)|mei|jun(i)?|jul(i)?|aug(ustus|\.)?|sep(t|t\.|tember|\.)?|okt(ober|\.)?|nov(ember|\.)?|dec(ember|\.)?)",
    "day_month": r"(\d{1,2}) (jan(uari|.)?|feb(ruari|\.)?|maart|mrt|apr(il|\.)|mei|jun(i)?|jul(i)?|aug(ustus|\.)?|sep(t|t\.|tember|\.)?|okt(ober|\.)?|nov(ember|\.)?|dec(ember|\.)?)",
    "year": r"(\d{4})",
    "time": r"\b(\d{1,2}|een|twee|drie|vier|vijf|zes|zeven|acht|negen|tien|elf|twaalf) (uu?ren)\b",
    "day_part": r"\b's (avonds|middags|ochtends)",
}

dutch_place_name_patterns = {
    "name_pattern": r"([A-Z](\w|-)+)",
}

dutch_place_name_patterns["in_placename"] = r"(in|tot) " + dutch_place_name_patterns["name_pattern"]

dutch_date_patterns["weekday_comma_day_month"] = dutch_date_patterns["week_day"] + "(,? (de|den)?) " + \
                                                 dutch_date_patterns["day_month"]

dutch_person_name_patterns["name_and_name_pattern"] = dutch_person_name_patterns["name_pattern"] + " en " + \
                                                      dutch_person_name_patterns["name_pattern"]
dutch_person_name_patterns["name_sequence_pattern"] = dutch_person_name_patterns["name_comma_pattern"] + \
                                                      dutch_person_name_patterns["name_and_name_pattern"]

# These pattern definitions specify:
# 1. which regex patterns to use for which pattern names
# 2. which regex match groups should be extracted
pattern_definitions = {
    "name": {
        "pattern": dutch_person_name_patterns["name_pattern"],
        "group_indices": [1],
        "type": "dutch_person_name",
    },
    "name_and_name": {
        "pattern": dutch_person_name_patterns["name_and_name_pattern"],
        "group_indices": [1, 10],
        "type": "dutch_person_name",
    },
    "name_sequence": {
        "pattern": dutch_person_name_patterns["name_sequence_pattern"],
        "group_indices": [1, 10, 19],
        "type": "dutch_person_name",
    },
    # commented out because it's not useful on it own:
    # "name_comma_name": {
    #    "pattern": dutch_person_name_patterns["name_comma_name_pattern"],
    #    "group_indices": [1, 10, 19],
    #    "type": "dutch_person_name",
    # },
    "weekday_comma_day_month": {
        "pattern": dutch_date_patterns["weekday_comma_day_month"],
        "group_indices": [1, 4, 5],
        "type": "dutch_date",
    },
}


def list_context_pattern_types(context_type=None):
    if not context_type:
        context_type = "all"
    if context_type not in context_pattern:
        print("ERROR - Unknown context type. Pick from:")
        for context_type in context_pattern:
            print("\t{t}".format(context_type))
        raise KeyError("Unknown context type")
    return [pattern_type for pattern_type in context_pattern[context_type]]


def list_pattern_names(name_only=True, pattern_type=None):
    if pattern_type:
        return [pattern_name for pattern_name in pattern_definitions if
                pattern_definitions[pattern_name]["type"] == pattern_type]
    else:
        return [pattern_name for pattern_name in pattern_definitions]


def list_pattern_definitions(pattern_type=None):
    if pattern_type:
        return [pattern_definitions[pattern_name] for pattern_name in pattern_definitions if
                pattern_definitions[pattern_name]["type"] == pattern_type]
    else:
        return pattern_definitions


def pattern_comma_then_context(name, pattern_definition, context_string):
    return {
        "name": name + "_comma_then_context",
        "pattern": pattern_definition["pattern"] + " ?, ?" + context_string,
        "group_indices": pattern_definition["group_indices"],
    }


def context_then_pattern(name, pattern_definition, context_string):
    return {
        "name": "context_then_" + name,
        "pattern": context_string + ",? " + pattern_definition["pattern"],
        "group_indices": pattern_definition["group_indices"],
    }


def pattern_before_context(name, pattern_definition, context_string, max_distance=10):
    return {
        "name": name + "_before_context",
        "pattern": pattern_definition["pattern"] + ".{d}".format(d=max_distance) + context_string,
        "group_indices": pattern_definition["group_indices"],
    }


def context_before_pattern(name, pattern_definition, context_string, max_distance=10):
    return {
        "name": "context_before_" + name,
        "pattern": context_string + ".{d}".format(d=max_distance) + pattern_definition["pattern"],
        "group_indices": pattern_definition["group_indices"],
    }


def get_search_patterns(pattern_type=None):
    if pattern_type:
        return {pattern_name: pattern_definitions[pattern_name] for pattern_name in pattern_definitions if
                pattern_definitions[pattern_name]["type"] == pattern_type}
    else:
        return pattern_definitions


def get_context_patterns(context_type: Union[None, str] = None) -> dict:
    if not context_type:
        context_type = "all"
    if context_type not in context_pattern:
        print("ERROR - Unknown context type. Pick from:")
        for context_type in context_pattern:
            print(f"\t{context_type}")
        raise KeyError("Unknown context type")
    return context_pattern[context_type]


# dictionary mapping regex pattern name to function to create context-specific regex pattern
context_pattern = {
    "person_name": {
        "pattern_comma_then_context": pattern_comma_then_context,
        "context_then_pattern": context_then_pattern,
    },
    "distance": {
        "pattern_before_context": pattern_before_context,
        "context_before_pattern": context_before_pattern,
    },
    "all": {
        "pattern_comma_then_context": pattern_comma_then_context,
        "context_then_pattern": context_then_pattern,
        "pattern_before_context": pattern_before_context,
        "context_before_pattern": context_before_pattern,
    },
}


def escape_string(string):
    string = string.replace("\\", r"\\").replace("/", r"\/")
    string = string.replace("[", r"\[").replace("]", r"\]").replace("(", r"\(").replace(")", r"\)")
    string = string.replace("{", r"\{").replace("}", r"\}")
    string = string.replace("*", r"\*").replace("?", r"\?").replace("+", r"\+")
    string = string.replace(".", r"\.").replace("|", r"\|")
    string = string.replace("!", r"\!").replace("^", r"\^").replace("$", r"\$")
    return string


def make_search_context_patterns(context_string, pattern_names, context_patterns):
    context_string = escape_string(context_string)
    patterns = []
    for context_pattern in context_patterns:
        for pattern_name in pattern_names:
            patterns += [
                context_patterns[context_pattern](pattern_name, pattern_definitions[pattern_name], context_string)]
    return patterns


if __name__ == "__main__":
    pattern_names = ["name_comma_name"]
    pattern_types = ["pattern_comma_then_context"]
    context_string = "Makelaar"

    make_search_context_patterns(context_string, pattern_names, pattern_types)



from typing import Dict, List, Set, Union
import copy
import string
import re
from collections import defaultdict

from fuzzy_search.fuzzy_phrase_model import PhraseModel
from fuzzy_search.fuzzy_match import PhraseMatch, Candidate, adjust_match_offsets
from fuzzy_search.fuzzy_phrase import Phrase
from fuzzy_search.fuzzy_string import text2skipgrams, SkipGram, score_levenshtein_similarity_ratio


default_config = {
    # these thresholds work when there are quite a few OCR errors
    # use higher thresholds for higher quality OCR to avoid
    # spurious matches.
    "char_match_threshold": 0.6,
    "ngram_threshold": 0.5,
    "levenshtein_threshold": 0.6,
    "skipgram_threshold": 0.2,
    # Is upper/lowercase a meaningful signal?
    "ignorecase": False,
    # should matches follow word boundaries?
    "use_word_boundaries": False,
    # for phrases that have variant phrasings
    "include_variants": False,
    # avoid matching with similar but different phrases
    "filter_distractors": False,
    # matching string can be lower/shorter than prhase
    "max_length_variance": 1,
    # higher ngram size allows fewer character differences
    "ngram_size": 2,
    # fewer skips is much faster but less exhaustive
    "skip_size": 2,
    # first check for exact matches to speed up fuzzy search
    "skip_exact_matching": False,
    # allow matches of partially overlapping phrase
    "allow_overlapping_matches": False,
    # the set of symbols to use as punctuation (for word boundaries)
    "punctuation": string.punctuation
}


class SkipMatches:

    def __init__(self, ngram_size: int, skip_size: int):
        self.ngram_size = ngram_size
        self.skip_size = skip_size
        self.skip_length = ngram_size + skip_size
        self.match_set: Dict[Phrase, set] = defaultdict(set)
        self.match_offsets = defaultdict(list)
        self.match_skipgrams: Dict[Phrase, List[SkipGram]] = defaultdict(list)
        self.phrases: Set[Phrase] = set()

    def add_skip_match(self, skipgram: SkipGram, phrase: Phrase) -> None:
        """Add a skipgram from a text that matches a phrase.

        :param skipgram: a skipgram from a text
        :type skipgram: SkipGram
        :param phrase: a phrase object that matches the skipgram
        :type phrase: Phrase
        """
        # track which skip grams of phrase are found in text
        self.match_set[phrase].add(skipgram.string)
        # track which text offsets match phrase skipgram
        self.match_offsets[phrase].append(skipgram.offset)
        self.match_skipgrams[phrase].append(skipgram)
        self.phrases.add(phrase)


def get_skipset_overlap(phrase: Phrase, skip_matches: SkipMatches) -> float:
    """Calculate the overlap between the set of skipgrams of a text and the skipgrams of a phrase.

    :param phrase: a phrase object that has been matched against a text
    :type phrase: Phrase
    :param skip_matches: a SkipMatches object containing the skipgram matches between a text and a list of phrases
    :type skip_matches: SkipMatches
    :return: the fraction of skipgrams in the phrase that overlaps with the text
    :rtype: float
    """
    return len(skip_matches.match_set[phrase]) / len(phrase.skipgram_set)


def filter_skipgram_threshold(skip_matches: SkipMatches, skip_threshold: float) -> List[Phrase]:
    """Filter the skipgram matches based on the skipgram overlap threshold.

    :param skip_matches: the phrases that matches the text
    :type skip_matches: SkipMatches
    :param skip_threshold: the threshold for the skipgram overlap between a text and a phrase
    :type skip_threshold: float
    :return: the list of phrases with a skipgram overlap that meets the threshold
    :rtype: List[Phrase]
    """
    return [phrase for phrase in skip_matches.phrases if get_skipset_overlap(phrase, skip_matches) >= skip_threshold]


def filter_overlapping_phrase_candidates(phrase_candidates: List[Candidate]) -> List[Candidate]:
    filtered: List[Candidate] = []
    if len(phrase_candidates) < 2:
        return phrase_candidates
    phrase_candidates.sort(key=lambda x: x.match_start_offset)
    prev_candidate = phrase_candidates[0]
    prev_score = score_levenshtein_similarity_ratio(prev_candidate.phrase.phrase_string, prev_candidate.match_string)
    for ci, curr_candidate in enumerate(phrase_candidates[1:]):
        if curr_candidate.match_end_offset > prev_candidate.match_start_offset:
            if curr_candidate.match_start_offset < prev_candidate.match_end_offset:
                # this candidate overlaps with the previous one, pick the best
                curr_score = score_levenshtein_similarity_ratio(curr_candidate.phrase.phrase_string,
                                                                curr_candidate.match_string)
                if curr_score > prev_score:
                    # this candidate is better, so skip the previous candidate
                    prev_candidate = curr_candidate
                    prev_score = curr_score
                elif curr_score == prev_score and len(curr_candidate.match_string) > len(prev_candidate.match_string):
                    # this candidate is longer with the same score, so skip the previous candidate
                    prev_candidate = curr_candidate
                    prev_score = curr_score
            else:
                # the previous candidate does not overlap with the current, so add it to filtered
                filtered.append(prev_candidate)
                prev_candidate = curr_candidate
                prev_score = score_levenshtein_similarity_ratio(curr_candidate.phrase.phrase_string,
                                                                curr_candidate.match_string)

    if len(filtered) == 0 or prev_candidate != filtered[-1]:
        filtered.append(prev_candidate)
    return filtered


def get_skipmatch_phrase_candidates(text: Dict[str, any], phrase: Phrase, skip_matches: SkipMatches,
                                    skipgram_threshold: float, max_length_variance: int = 1) -> List[Candidate]:
    """Find all candidate matches for a given phrase and SkipMatches object.

    :param text: the text object to match with phrases
    :type text: Dict[str, any]
    :param phrase: a phrase to find candidate matches for
    :type phrase: Phrase
    :param skip_matches: a Skipmatches object with matches between a text and a list of phrases
    :type skip_matches: SkipMatches
    :param skipgram_threshold: a threshold for how many skipgrams should match between a phrase and a candidate
    :type skipgram_threshold: float
    :param max_length_variance: the maximum difference in length between candidate and phrase
    :type max_length_variance: int
    :return: a list of candidate matches
    :rtype: List[Candidate]
    """
    candidates: List[Candidate] = []
    candidate = Candidate(phrase, max_length_variance=max_length_variance)
    last_index = len(skip_matches.match_offsets[phrase]) - 1
    # print(f"finding candidates for phrase ({len(phrase.phrase_string)}):", phrase.phrase_string)
    for ci, curr_offset in enumerate(skip_matches.match_offsets[phrase]):
        next_offset = None if ci == last_index else skip_matches.match_offsets[phrase][ci + 1]
        # print(ci, 'curr offset:', curr_offset, '\tskip:',
        #       skip_matches.match_skipgrams[phrase][ci].string, '\tnext offset:', next_offset)
        # add current skipgram to the candidate
        candidate.add_skip_match(skip_matches.match_skipgrams[phrase][ci])
        # if abs(candidate.skip_match_length() - len(candidate.phrase.phrase_string)) < max_length_variance:
        #     skip = skip_matches.match_skipgrams[phrase][ci]
        #     print(ci, curr_offset, "adding skip match:", skip.string, skip.offset, skip.length)
        #     print("candidate skips:", [skip.string for skip in candidate.skipgram_list],
        #           candidate.skip_match_length())
        #     print(candidate.get_skip_set_overlap(), candidate.get_match_string(text))
        # check if the current candidate is a potential match for the phrase
        if candidate.is_match(skipgram_threshold):
            candidate.match_string = candidate.get_match_string(text)
            # print("meets threshold:", candidate.match_string)
            # if this candidate has enough skipgram overlap, yield it as a candidate match
            if len(candidates) == 0 or not candidate.same_candidate(candidates[-1]):
                candidates.append(copy.deepcopy(candidate))
            if candidate.shift_start_skip():
                # candidate string is longer than phrase string check if shifting the start creates
                # a better candidate and if so, add that as well
                candidate.match_string = candidate.get_match_string(text)
                candidates.append(copy.deepcopy(candidate))
        if next_offset and next_offset - curr_offset > skip_matches.ngram_size + skip_matches.skip_size + 1:
            # if the gap between the current skipgram and the next is larger than an entire skipgram
            # the next skipgram does not belong to this candidate
            # start a new candidate for the next skipgram
            # print('curr_offset:', curr_offset, '\tnext_offset:', next_offset)
            # print('starting a new candidate')
            candidate = Candidate(phrase)
    # end of skipgrams reached, check if remaining candidate is a match
    # print('checking if final candidate is match')
    if candidate.is_match(skipgram_threshold):
        if len(candidates) == 0 or not candidate.same_candidate(candidates[-1]):
            candidate.match_string = candidate.get_match_string(text)
            candidates.append(copy.deepcopy(candidate))
        if candidate.shift_start_skip():
            # candidate string is longer than phrase string check if shifting the start creates
            # a better candidate and if so, add that as well
            candidate.match_string = candidate.get_match_string(text)
            candidates.append(copy.deepcopy(candidate))
    return candidates


def get_skipmatch_candidates(text: Dict[str, any], skip_matches: SkipMatches,
                             skipgram_threshold: float, phrase_model: PhraseModel,
                             max_length_variance: int = 1) -> List[Candidate]:
    """Find all candidate matches for the phrases in a SkipMatches object.

    :param text: the text object to match with phrases
    :type text: Dict[str, any]
    :param skip_matches: a SkipMatches object with matches between a text and a list of phrases
    :type skip_matches: SkipMatches
    :param skipgram_threshold: a threshold for how many skipgrams should match between a phrase and a candidate
    :type skipgram_threshold: float
    :param phrase_model: a phrase model, either as dictionary or as PhraseModel object
    :type phrase_model: PhraseModel
    :param max_length_variance: the maximum difference in length between candidate and phrase
    :type max_length_variance: int
    :return: a list of candidate matches
    :rtype: List[Candidate]
    """
    phrase_candidates = defaultdict(list)
    candidates: List[Candidate] = []
    for phrase in skip_matches.phrases:
        # print("get_skipmatch_candidates - phrase:", phrase.phrase_string)
        if get_skipset_overlap(phrase, skip_matches) < skipgram_threshold:
            continue
        if phrase.phrase_string in phrase_model.is_variant_of:
            match_phrase = phrase_model.is_variant_of[phrase.phrase_string]
        else:
            match_phrase = phrase.phrase_string
        phrase_candidates[match_phrase] += get_skipmatch_phrase_candidates(text, phrase, skip_matches,
                                                                           skipgram_threshold,
                                                                           max_length_variance=max_length_variance)
    for phrase_string in phrase_candidates:
        # print("phrase_candidates:", len(phrase_candidates[phrase_string]))
        filtered_candidates = filter_overlapping_phrase_candidates(phrase_candidates[phrase_string])
        # print(phrase_candidates)
        # print("filtered_candidates:", len(filtered_candidates))
        # for candidate in filtered_candidates:
        #     print(candidate.match_string, candidate.match_start_offset, candidate.match_end_offset)
        candidates += filtered_candidates
    return candidates


def get_text_dict(text: Union[str, dict], ignorecase: bool = False) -> dict:
    """Check that text is in a dictionary with an id property, so that passing a long text
    goes by reference instead of copying the long text string.

    :param text: a text string or text dictionary
    :type text: Union[str, dict]
    :param ignorecase: boolean flag for whether to ignore case
    :type ignorecase: bool
    :return: a text dictionary with an id property
    :rtype: dict
    """
    if isinstance(text, str):
        text = {"text": text, "id": None}
    if ignorecase:
        text["text"] = text["text"].lower()
    if "id" not in text:
        text["id"] = None
    return text


def candidates_to_matches(candidates: List[Candidate], text: dict, phrase_model: PhraseModel) -> List[PhraseMatch]:
    matches: List[PhraseMatch] = []
    for candidate in candidates:
        if candidate.phrase.phrase_string in phrase_model.is_variant_of:
            match_phrase_string = phrase_model.is_variant_of[candidate.phrase.phrase_string]
            match_phrase = phrase_model.phrase_index[match_phrase_string]
        else:
            match_phrase = candidate.phrase
        match = PhraseMatch(match_phrase, candidate.phrase,
                            candidate.match_string, candidate.match_start_offset, text["id"])
        match.add_scores(skipgram_overlap=candidate.get_skip_count_overlap())
        matches.append(match)
    return matches


def filter_matches_by_overlap(filtered_matches: List[PhraseMatch]) -> List[PhraseMatch]:
    sorted_matches = sorted(filtered_matches, key=lambda x: (x.offset, len(x.string)))
    filtered_matches = []
    overlapping = defaultdict(list)
    for match in sorted_matches:
        overlapping[(match.offset, len(match.string))].append(match)
    for offset_length in overlapping:
        if len(overlapping[offset_length]) == 1:
            filtered_matches.extend(overlapping[offset_length])
        else:
            best = max(overlapping[offset_length], key=lambda item: item.levenshtein_similarity)
            filtered_matches.append(best)
    return filtered_matches


class FuzzyPhraseSearcher(object):

    def __init__(self, config: Union[None, Dict[str, Union[str, int, float]]] = None):
        # default configuration
        self.char_match_threshold = 0.5
        self.ngram_threshold = 0.5
        self.skipgram_threshold = 0.2
        self.levenshtein_threshold = 0.5
        self.max_length_variance = 1
        self.allow_overlapping_matches = False
        self.skip_exact_matching = False
        self.use_word_boundaries = True
        self.ignorecase = False
        self.known_candidates = defaultdict(dict)
        self.distractor_terms = defaultdict(list)
        self.ngram_size = 2
        self.skipgram_index = defaultdict(set)
        self.early_skipgram_index = defaultdict(set)
        self.late_skipgram_index = defaultdict(set)
        self.skip_size = 2
        self.variant_map = defaultdict(dict)
        self.has_variant = defaultdict(dict)
        self.variant_skipgram_index = defaultdict(set)
        self.variant_early_skipgram_index = defaultdict(set)
        self.variant_late_skipgram_index = defaultdict(set)
        self.include_variants = False
        self.filter_distractors = False
        self.distractor_map = defaultdict(dict)
        self.has_distractor = defaultdict(dict)
        self.distractor_skipgram_index = defaultdict(set)
        self.distractor_early_skipgram_index = defaultdict(set)
        self.distractor_late_skipgram_index = defaultdict(set)
        self.phrases: Set[Phrase] = set()
        self.variants: Set[Phrase] = set()
        self.distractors: Set[Phrase] = set()
        self.phrase_model: Union[None, PhraseModel] = None
        self.debug = False
        self.punctuation = string.punctuation
        # non-default configuration
        self.config = copy.deepcopy(default_config)
        if config:
            self.configure(config)

    def configure(self, config: Dict[str, any]) -> None:
        """Configure the fuzzy searcher with a given config object.

        :param config: a config dictionary
        :type config: Dict[str, Union[str, int, float]]
        """
        if "char_match_threshold" in config:
            self.char_match_threshold = config["char_match_threshold"]
        if "ngram_threshold" in config:
            self.ngram_threshold = config["ngram_threshold"]
        if "skipgram_threshold" in config:
            self.skipgram_threshold = config["skipgram_threshold"]
        if "levenshtein_threshold" in config:
            self.levenshtein_threshold = config["levenshtein_threshold"]
        if "max_length_variance" in config:
            self.max_length_variance = config["max_length_variance"]
        if "use_word_boundaries" in config:
            self.use_word_boundaries = config["use_word_boundaries"]
        if "ignorecase" in config:
            self.ignorecase = config["ignorecase"]
        if "ngram_size" in config:
            self.ngram_size = config["ngram_size"]
        if "skip_size" in config:
            self.skip_size = config["skip_size"]
        if "include_variants" in config:
            self.include_variants = config["include_variants"]
        if "filter_distractors" in config:
            self.filter_distractors = config["filter_distractors"]
        if "allow_overlapping_matches" in config:
            self.allow_overlapping_matches = config["allow_overlapping_matches"]
        if "punctuation" in config:
            self.punctuation = config["punctuation"]
        if "debug" in config:
            self.debug = config["debug"]

    def index_phrase_model(self, phrase_model: Union[List[Dict[str, Union[str, int, float, list]]], PhraseModel]):
        """Add a phrase model to search for phrases in texts.

        :param phrase_model: a phrase model, either as dictionary or as PhraseModel object
        :type phrase_model: Union[List[Dict[str, Union[str, int, float, list]]], PhraseModel]
        """
        if isinstance(phrase_model, list):
            phrase_model = PhraseModel(model=phrase_model, config=self.config)
        self.phrase_model = phrase_model
        self.index_phrases(list(phrase_model.phrase_index.values()))
        self.index_variants(list(phrase_model.variant_index.values()))
        self.index_distractors(list(phrase_model.distractor_index.values()))

    def index_phrases(self, phrases: List[Union[str, Phrase]]) -> None:
        """Add a list of phrases to search for in texts.

        :param phrases: a list of phrases, either as string or as Phrase objects
        :type phrases: List[Union[str, Phrase]]
        """
        for phrase in phrases:
            if isinstance(phrase, str):
                phrase = Phrase(phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if phrase.ngram_size != self.ngram_size:
                searcher_size = f"{self.__class__.__name__} ({self.ngram_size}"
                raise ValueError(f"phrase has different ngram_size ({phrase.ngram_size}) than {searcher_size}")
            if phrase.skip_size != self.skip_size:
                searcher_size = f"{self.__class__.__name__} ({self.skip_size}"
                raise ValueError(f"phrase has different skip_size ({phrase.skip_size}) than {searcher_size}")
            self.phrases.add(phrase)
            if self.ignorecase:
                for skipgram in phrase.skipgrams_lower:
                    self.skipgram_index[skipgram.string].add(phrase)
                for skipgram_string in phrase.early_skipgram_index_lower:
                    self.early_skipgram_index[skipgram_string].add(phrase)
                for skipgram_string in phrase.late_skipgram_index_lower:
                    self.late_skipgram_index[skipgram_string].add(phrase)
            else:
                for skipgram in phrase.skipgrams:
                    self.skipgram_index[skipgram.string].add(phrase)
                for skipgram_string in phrase.early_skipgram_index:
                    self.early_skipgram_index[skipgram_string].add(phrase)
                for skipgram_string in phrase.late_skipgram_index:
                    self.late_skipgram_index[skipgram_string].add(phrase)
        if self.phrase_model is None:
            self.phrase_model = PhraseModel(phrases=list(self.phrases))

    def index_variants(self, variants: List[Union[str, Phrase]]) -> None:
        """Add a list of variant phrases to search for in texts.

        :param variants: a list of variants, either as string or as Phrase objects
        :type variants: List[Union[str, Phrase]]
        """
        for variant in variants:
            if isinstance(variant, str):
                variant = Phrase(variant, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if variant.ngram_size != self.ngram_size:
                searcher_size = f"{self.__class__.__name__} ({self.ngram_size}"
                raise ValueError(f"variant has different ngram_size ({variant.ngram_size}) than {searcher_size}")
            if variant.skip_size != self.skip_size:
                searcher_size = f"{self.__class__.__name__} ({self.skip_size}"
                raise ValueError(f"variant has different skip_size ({variant.skip_size}) than {searcher_size}")
            self.variants.add(variant)
            if self.ignorecase:
                for skipgram in variant.skipgrams_lower:
                    self.variant_skipgram_index[skipgram.string].add(variant)
                for skipgram_string in variant.early_skipgram_index_lower:
                    self.variant_early_skipgram_index[skipgram_string].add(variant)
                for skipgram_string in variant.late_skipgram_index_lower:
                    self.variant_late_skipgram_index[skipgram_string].add(variant)
            else:
                for skipgram in variant.skipgrams:
                    self.variant_skipgram_index[skipgram.string].add(variant)
                for skipgram_string in variant.early_skipgram_index:
                    self.variant_early_skipgram_index[skipgram_string].add(variant)
                for skipgram_string in variant.late_skipgram_index:
                    self.variant_late_skipgram_index[skipgram_string].add(variant)

    def index_distractors(self, distractors: List[Union[str, Phrase]]) -> None:
        """Add a list of distractor phrases to filter out likely incorrect phrase matches.

        :param distractors: a list of distractors, either as string or as Phrase objects
        :type distractors: List[Union[str, Phrase]]
        """
        for distractor in distractors:
            if isinstance(distractor, str):
                distractor = Phrase(distractor, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if distractor.ngram_size != self.ngram_size:
                searcher_size = f"{self.__class__.__name__} ({self.ngram_size}"
                raise ValueError(f"distractor has different ngram_size ({distractor.ngram_size}) than {searcher_size}")
            if distractor.skip_size != self.skip_size:
                searcher_size = f"{self.__class__.__name__} ({self.skip_size}"
                raise ValueError(f"distractor has different skip_size ({distractor.skip_size}) than {searcher_size}")
            self.distractors.add(distractor)
            if self.ignorecase:
                for skipgram in distractor.skipgrams_lower:
                    self.distractor_skipgram_index[skipgram.string].add(distractor)
                for skipgram_string in distractor.early_skipgram_index_lower:
                    self.distractor_early_skipgram_index[skipgram_string].add(distractor)
                for skipgram_string in distractor.late_skipgram_index_lower:
                    self.distractor_late_skipgram_index[skipgram_string].add(distractor)
            else:
                for skipgram in distractor.skipgrams:
                    self.distractor_skipgram_index[skipgram.string].add(distractor)
                for skipgram_string in distractor.early_skipgram_index:
                    self.distractor_early_skipgram_index[skipgram_string].add(distractor)
                for skipgram_string in distractor.late_skipgram_index:
                    self.distractor_late_skipgram_index[skipgram_string].add(distractor)

    def find_skipgram_matches(self, text: Dict[str, Union[str, int, float, list]],
                              include_variants: Union[None, bool] = None,
                              known_word_offset: Dict[int, Dict[str, any]] = None) -> SkipMatches:
        """Find all skipgram matches between text and phrases.

        :param text: the text object to match with phrases
        :type text: Dict[str, Union[str, int, float, list]]
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: bool
        :param known_word_offset: a dictionary of known words and their text offsets based on exact matches
        :type known_word_offset: Dict[int, Dict[str, any]]
        :return: a SkipMatches object contain all skipgram matches
        :rtype: SkipMatches
        """
        # skipmatch_count = 0
        known_word = None
        if include_variants is None:
            include_variants = self.include_variants
        if known_word_offset is None:
            known_word_offset = {}
        # print(known_word_offset)
        skip_matches = SkipMatches(self.ngram_size, self.skip_size)
        for skipgram in text2skipgrams(text["text"], self.ngram_size, self.skip_size):
            # print(skipgram.offset, skipgram.string)
            # print("skipgram:", skipgram.string)
            if skipgram.offset in known_word_offset:
                known_word = known_word_offset[skipgram.offset]
                # print("known word offset reached:", known_word)
            if known_word and skipgram.offset == known_word["end"]:
                # print("end of known word offset reached:", known_word)
                known_word = None
            """
            if known_word:
                for phrase_string in self.phrase_model.word_in_phrase[known_word["word"]]:
                    if phrase_string in known_word["match_phrases"]:
                        # this phrase was already found through exact match
                        continue
                    if phrase_string in self.phrase_model.phrase_index:
                        phrase = self.phrase_model.phrase_index[phrase_string]
                    elif include_variants and phrase_string in self.phrase_model.variant_index:
                        phrase = self.phrase_model.variant_index[phrase_string]
                    else:
                        # the phrase string either is a variant which is not include, or a distractor
                        continue
                    if skipgram.string in phrase.skipgrams:
                        skip_matches.add_skip_match(skipgram, phrase)
            else:
                for phrase in self.skipgram_index[skipgram.string]:
                    skip_matches.add_skip_match(skipgram, phrase)
                if include_variants:
                    for phrase in self.variant_skipgram_index[skipgram.string]:
                        skip_matches.add_skip_match(skipgram, phrase)
                """
            for phrase in self.skipgram_index[skipgram.string]:
                if phrase.max_offset > 0 and phrase.max_end < skipgram.offset + \
                        skipgram.length + self.max_length_variance:
                    # print(skipgram.offset, phrase.max_offset, phrase.max_end, phrase.phrase_string)
                    # print(f"skipping phrase {phrase.phrase_string} at offset", skipgram.offset)
                    continue
                if known_word:
                    if phrase.phrase_string not in self.phrase_model.word_in_phrase[known_word["word"]]:
                        # print("skipping phrase because doesn't match known word:", phrase.phrase_string)
                        continue
                    if phrase.phrase_string in known_word["match_phrases"]:
                        # print("skipping phrase because found as exact match:", phrase.phrase_string)
                        continue
                # print("\tphrase has skip:", phrase.phrase_string)
                # skipmatch_count += 1
                # print("adding skipmatch", skipmatch_count)
                skip_matches.add_skip_match(skipgram, phrase)
            if include_variants:
                for phrase in self.variant_skipgram_index[skipgram.string]:
                    if known_word:
                        if phrase.phrase_string not in self.phrase_model.word_in_phrase[known_word["word"]]:
                            # print("skipping phrase because doesn't match known word:", phrase.phrase_string)
                            continue
                        if phrase.phrase_string in known_word["match_phrases"]:
                            # print("skipping phrase because found as exact match:", phrase.phrase_string)
                            continue
                    skip_matches.add_skip_match(skipgram, phrase)
        # print("final skipmatch count:", skipmatch_count)
        return skip_matches

    def find_candidates(self, text: dict, use_word_boundaries: bool,
                        include_variants: Union[None, bool] = None,
                        known_word_offset: Dict[int, Dict[str, any]] = None) -> List[Candidate]:
        """Find candidate fuzzy matches for a given text.

        :param text: the text object to match with phrases
        :type text: dict
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: bool
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: bool
        :param known_word_offset: a dictionary of known words and their text offsets based on exact matches
        :type known_word_offset: Dict[int, Dict[str, any]]
        :return: a list of candidate matches
        :rtype: List[Candidate]
        """
        skip_matches = self.find_skipgram_matches(text, include_variants=include_variants,
                                                  known_word_offset=known_word_offset)
        candidates = get_skipmatch_candidates(text, skip_matches, self.skipgram_threshold, self.phrase_model,
                                              max_length_variance=self.max_length_variance)
        filtered = []
        use_word_boundaries = use_word_boundaries if use_word_boundaries is not None else self.use_word_boundaries
        for candidate in candidates:
            # print(candidate)
            if use_word_boundaries:
                # print("candidate:", candidate)
                adjusted_match = adjust_match_offsets(candidate.phrase.phrase_string, candidate.match_string,
                                                      text, candidate.match_start_offset, candidate.match_end_offset,
                                                      self.punctuation)
                # print("adjusted_match:", adjusted_match)
                if not adjusted_match:
                    continue
                candidate.match_start_offset = adjusted_match["match_start_offset"]
                candidate.match_end_offset = adjusted_match["match_end_offset"]
                candidate.match_string = adjusted_match["match_string"]
                # print("new match string:", candidate.match_string)
            filtered.append(candidate)
        return filtered

    def filter_matches_by_distractors(self, matches: List[PhraseMatch]) -> List[PhraseMatch]:
        filtered: List[PhraseMatch] = []
        for match in matches:
            if match.phrase.phrase_string in self.phrase_model.has_distractors:
                for distractor in self.phrase_model.has_distractors[match.phrase.phrase_string]:
                    score = score_levenshtein_similarity_ratio(match.string, distractor)
                    if score > match.levenshtein_similarity:
                        break
                else:
                    filtered.append(match)
            else:
                filtered.append(match)
        return filtered

    def filter_matches_by_threshold(self, matches: List[PhraseMatch]) -> List[PhraseMatch]:
        filtered: List[PhraseMatch] = []
        for match in matches:
            if match.character_overlap < self.char_match_threshold:
                continue
            if match.ngram_overlap < self.ngram_threshold:
                continue
            if match.levenshtein_similarity < self.levenshtein_threshold:
                continue
            filtered.append(match)
        return filtered

    def find_matches(self, text: Union[str, Dict[str, str]],
                     use_word_boundaries: Union[None, bool] = None,
                     allow_overlapping_matches: Union[None, bool] = None,
                     include_variants: Union[None, bool] = None,
                     filter_distractors: Union[None, bool] = None,
                     skip_exact_matching: bool = None) -> List[PhraseMatch]:
        """Find all fuzzy matching phrases for a given text. By default, a first pass of exact matching is conducted
        to find exact occurrences of phrases. This is to speed up the fuzzy matching pass

        :param text: the text (string or dictionary with 'text' property) to find fuzzy matching phrases in.
        :type text: Union[str, Dict[str, str]]
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: Union[None, bool]
        :param allow_overlapping_matches: boolean flag for whether to allow matches to overlap in their text ranges
        :type allow_overlapping_matches: Union[None, bool]
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: Union[None, bool]
        :param filter_distractors: boolean flag for whether to remove phrase matches that better match distractors
        :type filter_distractors: Union[None, bool]
        :param skip_exact_matching: boolean flag whether to skip the exact matching step
        :type skip_exact_matching: Union[None, bool]
        :return: a list of phrases matches
        :rtype: PhraseMatch
        """
        if self.phrase_model is None:
            raise ValueError("No phrase model indexed")
        text = get_text_dict(text, ignorecase=self.ignorecase)
        if use_word_boundaries is None:
            use_word_boundaries = self.use_word_boundaries
        if skip_exact_matching is None:
            skip_exact_matching = self.skip_exact_matching
        if not skip_exact_matching:
            # print("running exact matching")
            exact_matches = self.find_exact_matches(text, use_word_boundaries=use_word_boundaries,
                                                    include_variants=include_variants)
            known_word_offset = index_known_word_offsets(exact_matches)
        else:
            # print("skipping exact matching")
            exact_matches = []
            known_word_offset = {}
        # print('number of exact matches:', len(exact_matches))
        candidates = self.find_candidates(text, use_word_boundaries=use_word_boundaries,
                                          include_variants=include_variants, known_word_offset=known_word_offset)
        # print(candidates)
        matches = candidates_to_matches(candidates, text, self.phrase_model)
        # print(matches)
        filtered_matches = self.filter_matches_by_threshold(matches)
        if filter_distractors is None:
            filter_distractors = self.filter_distractors
        if filter_distractors:
            filtered_matches = self.filter_matches_by_distractors(filtered_matches)
        if allow_overlapping_matches is None:
            allow_overlapping_matches = self.allow_overlapping_matches
        if not allow_overlapping_matches:
            filtered_matches = filter_matches_by_overlap(filtered_matches)
        # print(exact_matches)
        # print(filtered_matches)
        selected_matches = filtered_matches + exact_matches
        return sorted(selected_matches, key=lambda x: x.offset)

    def find_exact_matches(self, text: Union[str, Dict[str, str]],
                           use_word_boundaries: Union[None, bool] = None,
                           include_variants: Union[None, bool] = None) -> List[PhraseMatch]:
        """Find all fuzzy matching phrases for a given text.

        :param text: the text (string or dictionary with 'text' property) to find fuzzy matching phrases in.
        :type text: Union[str, Dict[str, str]]
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: Union[None, bool]
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: Union[None, bool]
        :return: a list of phrases matches
        :rtype: PhraseMatch
        """
        exact_matches: List[PhraseMatch] = []
        text = get_text_dict(text, ignorecase=self.ignorecase)
        if use_word_boundaries is None:
            use_word_boundaries = self.use_word_boundaries
        if include_variants is None:
            include_variants = self.include_variants
        for exact_match in search_exact_phrases(self.phrase_model, text, use_word_boundaries=use_word_boundaries,
                                                include_variants=include_variants):
            exact_matches.append(exact_match)
        return exact_matches


def index_known_word_offsets(exact_matches: List[PhraseMatch]) -> Dict[int, Dict[str, any]]:
    exact_match_offset: Dict[int, Set[PhraseMatch]] = defaultdict(set)
    known_word_offset: Dict[int, Dict[str, any]] = defaultdict(dict)
    for exact_match in exact_matches:
        match_words = re.split(r"\W+", exact_match.string)
        # print("exact match:", exact_match)
        text_offset = exact_match.offset
        match_offset = 0
        # print("text_offset:", text_offset, "\tmatch_offset:", match_offset)
        for match_word in match_words:
            start = text_offset + match_offset + exact_match.string[match_offset:].index(match_word)
            # print("match_word:", match_word, "\tstart:", start)
            end = start + len(match_word)
            if start not in known_word_offset:
                known_word = {
                    "word": match_word,
                    "start": start,
                    "end": end,
                    "match_phrases": {exact_match.string}
                }
                # print(known_word)
                known_word_offset[start] = known_word
            known_word_offset[start]["match_phrases"].add(exact_match.string)
            match_word_offset = match_offset + exact_match.string[match_offset:].index(match_word)
            # print("text_offset:", text_offset, "\tmatch_offset:", match_offset)
        exact_match_offset[exact_match.offset].add(exact_match)
    return known_word_offset


def search_exact_phrases(phrase_model: PhraseModel, text: Dict[str, str],
                         ignorecase: bool = False, use_word_boundaries: bool = True,
                         include_variants: bool = False):
    if use_word_boundaries:
        # print('searching with word boundaries')
        return search_exact_phrases_with_word_boundaries(phrase_model, text, ignorecase=ignorecase,
                                                         include_variants=include_variants)
    else:
        return search_exact_phrases_without_word_boundaries(phrase_model, text, ignorecase=ignorecase,
                                                            include_variants=include_variants)


def add_exact_match_score(match: PhraseMatch) -> PhraseMatch:
    match.character_overlap = 1.0
    match.ngram_overlap = 1.0
    match.levenshtein_similarity = 1.0
    return match


def search_exact_phrases_with_word_boundaries(phrase_model: PhraseModel, text: Dict[str, str],
                                              ignorecase: bool = False, include_variants: bool = False):
    for word in re.finditer(r"\w+", text["text"]):
        if word.group(0) not in phrase_model.word_in_phrase:
            continue
        # print("\tword:", word)
        for phrase_string in phrase_model.first_word_in_phrase[word.group(0)]:
            phrase_word_offset = phrase_model.first_word_in_phrase[word.group(0)][phrase_string]
            phrase_start = word.start() - phrase_word_offset
            phrase_end = phrase_start + len(phrase_string)
            # print(phrase_start, phrase_end, phrase_string)
            if text["text"][phrase_start:phrase_end] == phrase_string:
                if phrase_start > 0 and re.match(r'\w', text["text"][phrase_start - 1]):
                    continue
                if phrase_end < len(text['text']) - 1 and re.match(r'\w', text['text'][phrase_end]):
                    continue
                if "phrase" in phrase_model.phrase_type[phrase_string]:
                    phrase = phrase_model.phrase_index[phrase_string]
                    match = PhraseMatch(phrase, phrase, phrase_string, phrase_start, text_id=text["id"])
                    yield add_exact_match_score(match)
                    # print("the matching phrase:", phrase)
                elif "variant" in phrase_model.phrase_type[phrase_string] and include_variants:
                    variant_phrase = phrase_model.variant_index[phrase_string]
                    main_phrase_string = phrase_model.is_variant_of[phrase_string]
                    main_phrase = phrase_model.phrase_index[main_phrase_string]
                    match = PhraseMatch(main_phrase, variant_phrase, phrase_string, phrase_start, text_id=text["id"])
                    yield add_exact_match_score(match)


def search_exact_phrases_without_word_boundaries(phrase_model: PhraseModel, text: Dict[str, str],
                                                 ignorecase: bool = False,
                                                 include_variants: bool = False):
    for phrase_string in phrase_model.phrase_index:
        phrase = phrase_model.phrase_index[phrase_string]
        for match in re.finditer(phrase.exact_string, text["text"]):
            phrase = phrase_model.phrase_index[phrase_string]
            match = PhraseMatch(phrase, phrase, phrase_string, match.start(), text_id=text["id"])
            yield add_exact_match_score(match)
    if include_variants:
        for phrase_string in phrase_model.variant_index:
            variant_phrase = phrase_model.variant_index[phrase_string]
            for match in re.finditer(variant_phrase.exact_string, text["text"]):
                variant_phrase = phrase_model.variant_index[phrase_string]
                main_phrase_string = phrase_model.is_variant_of[phrase_string]
                main_phrase = phrase_model.phrase_index[main_phrase_string]
                match = PhraseMatch(main_phrase, variant_phrase, phrase_string, match.start(), text_id=text["id"])
                yield add_exact_match_score(match)


def search_exact(phrase: Phrase, text: Dict[str, str], ignorecase: bool = False, use_word_boundaries: bool = True):
    search_string = phrase.extact_word_boundary_string if use_word_boundaries else phrase.exact_string
    if ignorecase:
        return re.finditer(search_string, text["text"], flags=re.IGNORECASE)
    else:
        return re.finditer(search_string, text["text"])


def get_known_word_offsets(match_ranges: List[Dict[str, any]], text_doc: Dict[str, str]) -> Dict[int, dict]:
    known_word_offset = {}
    offset = match_ranges[0]["s"]
    for match_range in match_ranges:
        print(match_range)
        print("offset:", offset)
        match_text = text_doc["text"][match_range["s"]:match_range["e"]]
        match_words = re.split(r"\W+", match_text)
        for match_word in match_words:
            start = offset + text_doc["text"][offset:].index(match_word)
            end = start + len(match_word)
            known_word = {
                "match_word": match_word,
                "start": start,
                "end": end,
                "match_phrases": set([phrase for phrase in match_range["phrases"]])
            }
            match_word_offset = offset + text_doc["text"][offset:].index(match_word)
            offset = match_word_offset + len(match_word)
            print(known_word)
            known_word_offset[known_word["start"]] = known_word
    return known_word_offset


def get_exact_match_ranges(exact_matches: List[PhraseMatch]) -> List[dict]:
    sorted_matches = sorted(exact_matches, key=lambda m: m.offset)
    match_ranges = []
    if len(exact_matches) == 0:
        return []
    first_match = exact_matches[0]
    match_range = {"s": first_match.offset, "e": first_match.end, "phrases": {first_match.phrase.phrase_string}}
    for mi, match in enumerate(sorted_matches[1:]):
        if match.offset > match_range["e"]:
            match_ranges.append(match_range)
            match_range = {"s": match.offset, "e": match.end, "phrases": set()}
        match_range["e"] = max(match_range["e"], match.end)
        match_range["phrases"].add(match.phrase.phrase_string)
        if match.phrase.phrase_string != match.variant.phrase_string:
            match_range["phrases"].add(match.variant.phrase_string)
    match_ranges.append(match_range)
    return match_ranges
